// MIT license (https://opensource.org/license/mit/)
// Initial Author: Guy Blelloch
// Developed as part of the flock library
// 
// A lock free (or lock-based) concurrent unordered_map using a hash
// table.   On a key type K and value type V it supports:
//
//   unordered_map<K,V,Hash=std::hash<K>,Equal=std::equal_to<K>>(n) :
//   constructor for table of size n
//
//   find(const K&) -> std::optional<V> : returns key if found, otherwise empty
//
//   insert(const K&, const V&) -> bool : if key not in the table it inserts the key
//   with the given value and returns true, otherwise does nothing and
//   returns false
//
//   remove(const K&) -> bool : if key is in the table it removes the entry
//   and returns true, otherwise it does nothing and returns false.
//
//   upsert(const K&, (const std::optional<V>&) -> V)) -> bool : if
//   key is in the table with value v then it calls the function (second argument)
//   with std::optional<V>(v), replaces the current value with the
//   returned value, and returns false.  Otherwise it calls the
//   function with std::nullopt and inserts the key into the table with the
//   returned value, and returns true.
//
//   size() -> long : returns the size of the table.  Not linearizable with
//   the other functions, and takes time proportional to the table size.
//
// Implementation: Each bucket points to a structure (Node) containing
// an array of entries.  Nodes come in varying sizes and on update the
// node is copied.  Allows arbitrary growing, but only efficient if
// not much larger than the original given size (i.e. number of
// buckets is fixed, but number of entries per bucket can grow).  Time
// is proportional to the size of the bucket.
//
// Define USE_LOCKS to use locks.  The lock-based version only
// acquires locks on updates.  Locks are faster with high contention
// workloads that include reads.  The lock-free version is marginally
// faster on low-contention uniform workloads, or if updates only.
// Also the lock-based version can suffer under oversubscription (more
// user threads than available hardware threads).
//
// The type for keys and values must be copyable, and might be copied
// by the hash_table even when not being updated (e.g. when another
// key in the same bucket is being updated). 
//
// The upsert operation takes a function f of type
//   (const std::optional<V>&) -> V
// If using locks, f is executed with no write-write races.
// There can be concurrent reads on the old value, hence the const to prevent
// any read-write races.

#ifndef PARLAYHASH_H_
#define PARLAYHASH_H_

#include <atomic>
#include <optional>
#include "epoch.h"
#include "lock.h"
//#define USE_LOCKS 1

namespace parlay {
  
template <typename K,
	  typename V,
	  class Hash = std::hash<K>,
	  class KeyEqual = std::equal_to<K>>
struct unordered_map {
private:
  struct KV {K key; V value;};

  template <typename Range>
  static int find_in_range(const Range& entries, long cnt, const K& k) {
    for (int i=0; i < cnt; i++)
      if (KeyEqual{}(entries[i].key, k)) return i;
    return -1;
  }

  // The following three functions copy a range and
  // insert/update/remove the specified key.  No ordering is assumed
  // within the range.  Insert assumes k does not appear, while
  // update/remove assume it does appear at index idx.
  template <typename Range, typename RangeIn>
  static void copy_and_insert(Range& out, const RangeIn& entries, long cnt,
			      const K& k, const V& v) {
    assert(cnt >= 0);
    for (int i=0; i < cnt; i++) out[i] = entries[i];
    out[cnt] = KV{k,v};
  }

  template <typename Range, typename RangeIn, typename F>
  static void copy_and_update(Range& out, const RangeIn& entries, long cnt,
			      const K& k, const F& f, long idx) {
    assert(cnt > 0);
    for (int i = 0; i < idx; i++) out[i] = entries[i];
    out[idx].key = entries[idx].key;
    out[idx].value = f(entries[idx].value);
    for (int i = idx+1; i < cnt; i++) out[i] = entries[i];
  }

  template <typename Range, typename RangeIn>
  static void copy_and_remove(Range& out, const RangeIn& entries, long cnt, const K& k, long idx) {
    assert(cnt > 1);
    for (int i = 0; i < idx; i++) out[i] = entries[i];
    for (int i = idx; i < cnt-1; i++) out[i] = entries[i+1];
  }
  
  // Each bucket points to a Node of some Size, or to a BigNode (defined below)
  // A node contains an array of up to Size entries (actual # of entries given by cnt)
  // Sizes are 1, 3, 7, 31
  template <int Size>
  struct Node {
    using node = Node<0>;
    int cnt;
    KV entries[Size];

    // return index of key in entries, or -1 if not found
    int find_index(const K& k) {
      if (cnt <= 31) return find_in_range(entries, cnt, k);
      else return find_in_range(((BigNode*) this)->entries, cnt, k);
    }

    // return optional value found in entries given a key
    std::optional<V> find(const K& k) {
      if (cnt <= 31) { // regular node
	if (KeyEqual{}(entries[0].key, k)) // shortcut for common case
	   return entries[0].value;
	int i = find_in_range(entries+1, cnt-1, k);
	if (i == -1) return {};
	else return entries[i].value;
      } else { // big node
	int i = find_in_range(((BigNode*) this)->entries, cnt, k);
	if (i == -1) return {};
	else return ((BigNode*) this)->entries[i].value;
      }
    }

    // copy and insert
    Node(node* old, const K& k, const V& v) {
      cnt = (old == nullptr) ? 1 : old->cnt + 1;
      copy_and_insert(entries, old->entries, cnt-1, k, v);
    }

    // copy and update
    template <typename F>
    Node(long idx, node* old, const K& k, const F& f) : cnt(old->cnt) {
      assert(old != nullptr);
      copy_and_update(entries, old->entries, cnt, k, f, idx);
    }

    // copy and remove
    Node(long idx, node* old, const K& k) : cnt(old->cnt - 1) {
      if (cnt == 31) copy_and_remove(entries, ((BigNode*) old)->entries, cnt+1, k, idx);
      else copy_and_remove(entries, old->entries, cnt+1, k, idx);
    }
  };
  using node = Node<0>;

  // If a node overflows (cnt > 31), then it becomes a big node and its content
  // is stored indirectly in an std::vector.
  struct BigNode {
    using entries_type = parlay::sequence<KV>;
    int cnt;
    entries_type entries;

    BigNode(node* old, const K& k, const V& v) : cnt(old->cnt + 1) {
      entries = parlay::tabulate(cnt, [] (long i) {return KV{};}, cnt);
      //entries = entries_type(cnt);
      if (old->cnt == 31) copy_and_insert(entries, old->entries, old->cnt, k, v);
      else copy_and_insert(entries, ((BigNode*) old)->entries, old->cnt, k, v);
    }

    template <typename F>
    BigNode(long idx, node* old, const K& k, const F& f) : cnt(old->cnt) {
      entries = parlay::tabulate(cnt, [] (long i) {return KV{};}, cnt);
      //entries = entries_type(cnt);
      copy_and_update(entries, ((BigNode*) old)->entries, cnt, k, f, idx);  }

    BigNode(long idx, node* old, const K& k) : cnt(old->cnt - 1) {
      entries = parlay::tabulate(cnt, [] (long i) {return KV{};}, cnt);
      //entries = entries_type(cnt);
      copy_and_remove(entries, ((BigNode*) old)->entries, cnt+1, k, idx); }
  };

  using Node1 = Node<1>;
  using Node3 = Node<3>;
  using Node7 = Node<7>;
  using Node31 = Node<31>;
  static parlay::memory_pool<Node1> node_pool_1;
  static parlay::memory_pool<Node3> node_pool_3;
  static parlay::memory_pool<Node7> node_pool_7;
  static parlay::memory_pool<Node31> node_pool_31;
  static parlay::memory_pool<BigNode> big_node_pool;

  static node* insert_to_node(node* old, const K& k, const V& v) {
    if (old == nullptr) return (node*) node_pool_1.new_obj(old, k, v);
    if (old->cnt < 3) return (node*) node_pool_3.new_obj(old, k, v);
    else if (old->cnt < 7) return (node*) node_pool_7.new_obj(old, k, v);
    else if (old->cnt < 31) return (node*) node_pool_31.new_obj(old, k, v);
    else return (node*) big_node_pool.new_obj(old, k, v);
  }

  template <typename F>
  static node* update_node(node* old, const K& k, const F& f, long idx) {
    assert(old != nullptr);
    if (old == nullptr) return (node*) node_pool_1.new_obj(idx, old, k, f);
    if (old->cnt < 3) return (node*) node_pool_3.new_obj(idx, old, k, f);
    else if (old->cnt < 7) return (node*) node_pool_7.new_obj(idx, old, k, f);
    else if (old->cnt < 31) return (node*) node_pool_31.new_obj(idx,old, k, f);
    else return (node*) big_node_pool.new_obj(idx, old, k, f);
  }

  static node* remove_from_node(node* old, const K& k, long idx) {
    assert(old != nullptr);
    if (old->cnt == 1) return (node*) nullptr;
    if (old->cnt == 2) return (node*) node_pool_1.new_obj(idx, old, k);
    else if (old->cnt <= 4) return (node*) node_pool_3.new_obj(idx, old, k);
    else if (old->cnt <= 8) return (node*) node_pool_7.new_obj(idx, old, k);
    else if (old->cnt <= 32) return (node*) node_pool_31.new_obj(idx, old, k);
    else return (node*) big_node_pool.new_obj(idx, old, k);
  }

  static void retire_node(node* old) {
    if (old == nullptr);
    else if (old->cnt == 1) node_pool_1.retire((Node1*) old);
    else if (old->cnt <= 3) node_pool_3.retire((Node3*) old);
    else if (old->cnt <= 7) node_pool_7.retire((Node7*) old);
    else if (old->cnt <= 31) node_pool_31.retire((Node31*) old);
    else big_node_pool.retire((BigNode*) old);
  }

  static void destruct_node(node* old) {
    if (old == nullptr);
    else if (old->cnt == 1) node_pool_1.destruct((Node1*) old);
    else if (old->cnt <= 3) node_pool_3.destruct((Node3*) old);
    else if (old->cnt <= 7) node_pool_7.destruct((Node7*) old);
    else if (old->cnt <= 31) node_pool_31.destruct((Node31*) old);
    else big_node_pool.destruct((BigNode*) old);
  }

  // *********************************************
  // The bucket and table structures
  // *********************************************

  using bucket = std::atomic<node*>;

  struct Table {
    parlay::sequence<bucket> table;
    size_t size;
    bucket* get_bucket(const K& k) {
      size_t idx = Hash{}(k)  & (size-1u);
      return &table[idx];
    }
    Table(size_t n) {
      int bits = 1 + parlay::log2_up(n);
      size = 1ul << bits;
      table = parlay::sequence<bucket>(size);
    }
  };

  Table hash_table;

  // *********************************************
  // The internal update functions (insert, upsert and remove)
  // *********************************************


  // try to install a new node in bucket s
  static std::optional<bool> try_update(bucket* s, node* old_node, node* new_node, bool ret_val) {
#ifndef USE_LOCKS
    if (s->load() == old_node &&
	s->compare_exchange_strong(old_node, new_node)) {
#else  // use try_lock
    if (locks.try_lock((long) s, [=] {
	    if (s->load() != old_node) return false;
	    *s = new_node;
	    return true;})) {
#endif
      retire_node(old_node);
      return ret_val;
    } 
    destruct_node(new_node);
    return {};
  }

  static std::optional<bool> try_insert_at(bucket* s, const K& k, const V& v) {
    node* old_node = s->load();
    if (old_node != nullptr && old_node->find_index(k) != -1) return false;
    return try_update(s, old_node, insert_to_node(old_node, k, v), true);
  }

  template <typename F>
  static std::optional<bool> try_upsert_at(bucket* s, const K& k, F& f) {
    node* old_node = s->load();
    long idx;
    bool found = old_node != nullptr && (idx = old_node->find_index(k)) != -1;
    if (!found)
      return try_update(s, old_node, insert_to_node(old_node, k, f(std::optional<V>())), true);
    else
#ifndef USE_LOCKS
      return try_update(s, old_node, update_node(old_node, k, f, idx), false);
#else  // use try_lock
    if (locks.try_lock((long) s, [=] {
        if (s->load() != old_node) return false;
	*s = update_node(old_node, k, f, idx); // f applied within lock
	return true;})) {
      retire_node(old_node);
      return false;
    } else return {};
#endif
  }

  static std::optional<bool> try_remove_at(bucket* s, const K& k) {
      node* old_node = s->load();
      long idx;
      if (old_node == nullptr || (idx = old_node->find_index(k)) == -1) return false;
      return try_update(s, old_node, remove_from_node(old_node, k, idx), true);
  }

public:
  // *********************************************
  // The public interface
  // *********************************************

  unordered_map(size_t n) : hash_table(Table(n)) {}
  ~unordered_map() {
    auto& table = hash_table.table;
    parlay::parallel_for (0, table.size(), [&] (size_t i) {
      retire_node(table[i].load());});
  }
  
  std::optional<V> find(const K& k) {
    bucket* s = hash_table.get_bucket(k);
    __builtin_prefetch (s);
    return parlay::with_epoch([&] {
      node* x = s->load();
      if (x == nullptr) return std::optional<V>();
      //if (KeyEqual{}(x->entries[0].key, k))
      //  return std::optional<V>(x->entries[0].value);
      return std::optional<V>(x->find(k));
    });
  }

  bool insert(const K& k, const V& v) {
    bucket* s = hash_table.get_bucket(k);
    __builtin_prefetch (s);
    return parlay::with_epoch([&] {
      return parlay::try_loop([&] {return try_insert_at(s, k, v);});});
  }

  template <typename F>
  bool upsert(const K& k, const F& f) {
    bucket* s = hash_table.get_bucket(k);
    __builtin_prefetch (s);
    return parlay::with_epoch([&] {
      return parlay::try_loop([&] {return try_upsert_at(s, k, f);});});
  }

  bool remove(const K& k) {
    bucket* s = hash_table.get_bucket(k);
    __builtin_prefetch (s);
    return parlay::with_epoch([&] {
      return parlay::try_loop([&] {return try_remove_at(s, k);});});
  }

  long size() {
    auto& table = hash_table.table;
    auto s = parlay::tabulate(table.size(), [&] (size_t i) {
	      node* x = table[i].load();
	      if (x == nullptr) return 0;
	      else return x->cnt;});
    return parlay::reduce(s);
  }
};

template <typename K, typename V, typename H, typename E>
parlay::memory_pool<typename unordered_map<K,V,H,E>::Node1> unordered_map<K,V,H,E>::node_pool_1;
template <typename K, typename V, typename H, typename E>
parlay::memory_pool<typename unordered_map<K,V,H,E>::Node3> unordered_map<K,V,H,E>::node_pool_3;
template <typename K, typename V, typename H, typename E>
parlay::memory_pool<typename unordered_map<K,V,H,E>::Node7> unordered_map<K,V,H,E>::node_pool_7;
template <typename K, typename V, typename H, typename E>
parlay::memory_pool<typename unordered_map<K,V,H,E>::Node31> unordered_map<K,V,H,E>::node_pool_31;
template <typename K, typename V, typename H, typename E>
parlay::memory_pool<typename unordered_map<K,V,H,E>::BigNode> unordered_map<K,V,H,E>::big_node_pool;

} // namespace parlay
#endif //PARLAYHASH_H_
